#!/usr/local/bin/tclsh8.4

#########################
# pftpshell.tcl v1.9    #
# leprechau@Efnet       #
# private..do not leak  #
#########################
# instructions:
#
# 0) chmod 0700 pftpshell (THIS IS CRITICAL)
#    chmod 0600 pftpshell.conf (THIS IS CRITICAL)
#
# 1) start server by: ./pftpshell
#    (note server runs in foreground)
#
# 2) configure mirc script with same authkey
#    and other settings (see mirc script for more info)
##

## settings ##
set PidFile "pftpshell.pid"
# server pid file (usefull if you want to crontab server)
set ListenPort "50000"
# listening port (mirc connects to this)
set MyAddr ""
# ip address to bind to (leave blank to listen on all interfaces)
set AuthKey "AHkjdf1hakljs34fAFH431jksadfsad4123fhLJFSAKDH72"
# authentication key (set to a long random string)
set IpRestrict "1"
# restriction: (0) server not ip restricted (1) server is restricted (use this)
set IpList "127.*.*.* 192.168.*.* 10.0.0.*"
# space deliminated list of accepted ip addresses (full wildcard masks allowed)
set PftpBin "./pftp"
# location of your pftp binary (no trailing /)
set LogType "2"
# logtype: (0) to log to stdout (1) log to logfile (2) log to both stdout and logfile
set LogFile "pftpshell.log"
# location and name of logfile (no trailing /)

###############################################
## begin script...DO NOT EDIT...begin script ##
###############################################

if {[file exists pftpshell.conf]} { source pftpshell.conf }

proc timestamp {} { return [clock format [clock seconds] -format "%a %b %d %H:%M:%S %Z %Y"] }

proc bgerror { text } {
global LogType
   if {(($LogType == 0) || ($LogType == 2))} {
      puts "$text"
   }
   if {(($LogType == 1) || ($LogType == 2))} {
      write_log "Error: $text"
   }
   exit 0
}

if {[file exists $PidFile]} {
   set pidfile [open $PidFile r]
   set ppid [gets $pidfile]
   close $pidfile
   unset pidfile
   if {[catch {set pslist "[exec ps -p $ppid]"} exec_error] == 0} {
      puts ""
      puts "Error: Already running? (pid: $ppid)"
      puts "       If not, please remove pidfile '$PidFile' and restart server."
      puts ""
      exit 0
   }
}

proc write_log { text } {
global LogType LogFile
   if {(($LogType == 0) || ($LogType == 2))} {
      puts "$text"
   }
   if {(($LogType == 1) || ($LogType == 2))} {
      set logfile [open $LogFile a+ 0600]
      puts $logfile "$text"
      close $logfile
   }
}

proc init {} {
global PidFile ListenPort MyAddr IpRestrict IpList ServerSock
   if {$MyAddr != ""} {
      set ServerSock [socket -server server_connect -myaddr $MyAddr $ListenPort]
   } else {
      set ServerSock [socket -server server_connect $ListenPort]
   }
   set pidfile [open "$PidFile" w]
   puts $pidfile "[pid]"
   close $pidfile
   puts ""
   puts "Pftp Input Server: Listening Port $ListenPort"
   if {$MyAddr != ""} {
      puts "Pftp Input Server: Listening Host $MyAddr"
   } else {
      puts "Pftp Input Server: Listening Host *"
   }
   if {$IpRestrict == "1"} {
      puts "Pftp Input Server: IP Restriction Active"
      puts "Allowed IPs      : $IpList"
   } else {
      puts "Pftp Input Server: IP Restriction NOT ACTIVE (please activate for increased security)"
   }
   puts "Pftp Input Server: Launched...Go thank leprechau@EFnet (pid: [pid])"
   puts "Pftp Input Server: Report any bugs\\comments to ahurt@anbcs.com"
   puts "Pftp Input Server: Press <CTRL> + <C> to close server."
   puts ""
   write_log "Server socket ($ServerSock) opened at [timestamp]"
   vwait forever
}

proc close_socket { sock } {
global Session
   close $sock
   catch {unset Session(addr,$sock)}
}

proc server_connect { sock addr cport } {
global Session IpRestrict IpList
   set allow_connect 0
   set Session(addr,$sock) [list $addr $cport]
   fconfigure $sock -buffering line
   if {$IpRestrict == 1} {
      foreach ip "$IpList" {
         if {[string match "$ip" "$addr"]} {
            set allow_connect 1
         }
      }
   } else { set allow_connect 1 }
   if {$allow_connect != 1} {
      puts $sock "Error, You are not authorized to use this service!"
      puts $sock "Closing socket: Attempt from $addr at [timestamp] logged!"
      close_socket $sock
      write_log "Closed unauthorized connect ($sock) from $addr at [timestamp]"
   } else {
      write_log "Accept $sock from $addr port $cport"
      fileevent $sock readable [list input_handler $sock $addr]
   }
}

proc input_handler { sock addr } {
global Session ServerSock MyAddr ListenPort PftpBin AuthKey
   if {[eof $sock] || [catch {gets $sock line}]} {
      close $sock
      write_log "Closed $Session(addr,$sock)"
      unset Session(addr,$sock)
   } else {
      if {![string equal "[lindex [split $line] 0]" "$AuthKey"]} { 
         puts $sock "Error, You are not authorized to use this service!"
         puts $sock "Closing socket: Attempt from $addr at [timestamp] logged!"
         close_socket $sock
         write_log "Closed unauthorized connect ($sock) from $addr at [timestamp]"
      } else {
         if {((![string equal "[lindex [split $line] 1]" "exit"]) && ([llength [split $line]] >= 4))} {
            set execargs "[lrange [split $line] 1 end]"
            write_log "Accepting input: $execargs"
            puts $sock "Thank You, please drive through."
            puts $sock "Executing: $PftpBin $execargs"
            puts $sock "Closing socket: Attempt from $addr at [timestamp] logged!"
            close_socket $sock
            write_log "Closed connection ($sock) from $addr at [timestamp]"
            write_log "Executing: $PftpBin $execargs"
            close $ServerSock
            write_log "Server socket ($ServerSock) closed at [timestamp]"
            eval exec $PftpBin $execargs >> [exec tty]
            if {$MyAddr != ""} {
               set ServerSock [socket -server server_connect -myaddr $MyAddr $ListenPort]
            } else {
               set ServerSock [socket -server server_connect $ListenPort]
            }
            puts "Pftp Input Server: Pftp closed, waiting for next connection..."
            puts "Pftp Input Server: Press <CTRL> + <C> to close server."
            write_log "Server socket ($ServerSock) opened at [timestamp]"
            vwait forever
         } elseif {[llength [split $line]] < 4} {
            puts $sock "Error, Invalid argument list: [lrange [split $line] 1 end]"
            puts $sock "Closing socket: Attempt from $addr at [timestamp] logged!"
            close_socket $sock
            write_log "Closed connection ($sock) from $addr at [timestamp]"
         } elseif {[string equal "[lindex [split $line] 1]" "exit"]} {
            puts $sock "Thank You, please drive through."
            close_socket $sock
         } else { return }
      }
   }
}
init
